/**
 * basic TCP/IP functions
 *
 * @author		Sven Skrabal
 * @copyright	(C) 2007 Sven Skrabal
 * @license		GNU General Public License (v3)
 *
 */

#include "stdafx.h"

/**
 * redirect the in and output of two sockets e.g.
 * socket 1 receive -> socket 2 send
 * socket 2 receive -> socket 1 send
 *
 * @param	sClientSocket	first client socket to redirect
 * @param	sServerSocket	second server socket to redirect
 * @return	returns -1 on error, a value of 0 or greater for success
 *
 */
int redirectSockets(SOCKET *sClientSocket, SOCKET *sServerSocket)
{
	char szTemp[16384];
	FD_SET fdSet;
	struct timeval tvTimeout;
	int iRetVal;

	//set timeout in structure
	tvTimeout.tv_sec = 0;
	tvTimeout.tv_usec = SELECT_TIMEOUT;

	while(true)
	{
		//for non blocking operation add sockets into check queue
		FD_ZERO(&fdSet);
		FD_SET(*(sClientSocket), &fdSet);
		FD_SET(*(sServerSocket), &fdSet);

		//check sockets for data
		if(select(0, &fdSet, NULL, NULL, &tvTimeout) == SOCKET_ERROR)
		{
			return -1;
		}

		//client socket has data
		if(FD_ISSET(*(sClientSocket), &fdSet))
		{
			//recieve data and send to server
			iRetVal = recv(*(sClientSocket), szTemp, 16384, 0);
			if(iRetVal == -1 || iRetVal == 0)return -1;

			iRetVal = send(*(sServerSocket), szTemp, iRetVal, 0);
			if(iRetVal == -1)return -1;
		}

		//server socket has data
		if(FD_ISSET(*(sServerSocket), &fdSet))
		{
			//recieve data and send to client
			iRetVal = recv(*(sServerSocket), szTemp, 16384, 0);
			if(iRetVal == -1 || iRetVal == 0)return -1;

			iRetVal = send(*(sClientSocket), szTemp, iRetVal, 0);
			if(iRetVal == -1)return -1;
		}
	};

	return 0;
}

/**
 * creates a socket and connects to given destination using a fixed timeout value
 *
 * @param	sSocket			pointer to socket to use
 * @param	siAddr			pointer to sockaddr structure to use
 * @param	szHostnameOrIp	null terminated string containing server to connect to
 * @param	iPort			destination port to connect to
 * @param	iNoneBlocking	option to set none blocking mode for that socket
 * @return	returns -1 on error, a value greater or equal than 0 for success
 *
 */
int createSocketAndConnect(SOCKET *sSocket, SOCKADDR_IN *siAddr, char *szHostnameOrIp, int iPort, int iNoneBlocking)
{
	unsigned long ulNonBlocking=2, ulBlocking=0;
	DWORD dwLastTick;
	int iRetVal;
	FD_SET fdSet;
	struct timeval tvTimeout;

	//set timeout in structure
	tvTimeout.tv_sec = 0;
	tvTimeout.tv_usec = SELECT_TIMEOUT;

	//create a socket
	*(sSocket) = socket(AF_INET, SOCK_STREAM, 0);

	if(*(sSocket) == INVALID_SOCKET)
	{
		return -1;
	}

	//set port and TCP/IP operation mode
	memset(siAddr, 0, sizeof(SOCKADDR_IN));
	((SOCKADDR_IN *)siAddr)->sin_family = AF_INET;
	((SOCKADDR_IN *)siAddr)->sin_port = htons(iPort);

	//resolve target server address
	if(resolveDNS(szHostnameOrIp, siAddr) == -1)
	{
		return -1;
	}

	//set to none blocking mode for connect timeouts
	if(ioctlsocket(*(sSocket), FIONBIO, &ulNonBlocking) != 0)
	{
			return -1;
	}

	//connect to destination and measure the time
	dwLastTick = GetTickCount();

	iRetVal = connect(*(sSocket), (SOCKADDR*)siAddr, sizeof(SOCKADDR));

	while(dwLastTick + CONNECT_TIMEOUT > GetTickCount())
	{
		FD_ZERO(&fdSet);
		FD_SET(*(sSocket), &fdSet);

		//check if socket is writeable/ready
		if(select(0, NULL, &fdSet, NULL, &tvTimeout) == SOCKET_ERROR)
		{			
			return -1;
		}

		//socket is ready and connection was successful
		if(FD_ISSET(*(sSocket), &fdSet))
		{
			//set socket to blocking mode, if requested
			if(iNoneBlocking == 0)
			{
				if(ioctlsocket(*(sSocket), FIONBIO, &ulBlocking)!=0)
				{
					return -1;
				}
			}
			
			return 0;
		}
	};

	return -1;
}

/**
 * resolves the ip of a hostname and transforms it to network byte order
 * resolved address will be stored in given structure for further processing
 *
 * @param	szHostnameOrIp		null terminated string containing ip or hostname
 * @param	siAddr				pointer to sockaddr structure to store resolved address
 * @return	returns -1 on error, a value greater or equal than 0 for success
 * 
 */
int resolveDNS(char *szHostnameOrIp, SOCKADDR_IN *siAddr)
{
	unsigned long ulIp;
	HOSTENT* hoHe;

	//is the string an ip?
	ulIp = inet_addr(szHostnameOrIp);

	if(ulIp != INADDR_NONE)
	{
		//just copy the data and return
		siAddr->sin_addr.s_addr = ulIp;
		return 0;
	}
	else
	{
		//resolve ip for hostname
		hoHe = gethostbyname(szHostnameOrIp);

		if(hoHe == NULL)
		{
			return -1;
		}
		else
		{
			//copy resolved ip
			memcpy(&(siAddr->sin_addr), hoHe->h_addr_list[0],4);
			return 0;
		}
	}
}

/**
 * creates a listen socket to create a server for incoming connections
 *
 * @param	sSocket		pointer to socket to use for server
 * @param	siAddr		address structure to use
 * @param	iPort		at which port should the socket listen on
 * @param	iReuseAddr	flag to determine if port should be reused (required after server crash or unclean socket shutdowns)
 * @return	returns -1 on error, a value greater or equal than 0 for success
 *
 */
int createListenSocket(SOCKET *sSocket, SOCKADDR_IN *siAddr, int iPort, int iNoneBlocking, int iReuseAddr)
{
	int iTemp=1;
	unsigned long ulNonBlocking=2;

	//create a socket to accept incoming connections
	*(sSocket) = socket(AF_INET, SOCK_STREAM, 0);

	if(*(sSocket) == INVALID_SOCKET)
	{
		return -1;
	}

	//set socket to none blocking mode
	if(iNoneBlocking == 1)
	{
		if(ioctlsocket(*(sSocket), FIONBIO, &ulNonBlocking)!=0)
		{
			return -1;
		}
	}

	//set socket to reuse the local address
	if(iReuseAddr == 1)
	{
		if(setsockopt(*(sSocket), SOL_SOCKET, SO_REUSEADDR, (const char*)&iTemp, sizeof(int)) == SOCKET_ERROR)
		{
			return -1;
		}
	}

	//set port and TCP/IP operation mode
	memset(siAddr, 0, sizeof(SOCKADDR_IN));
	((SOCKADDR_IN *)siAddr)->sin_family = AF_INET;
	((SOCKADDR_IN *)siAddr)->sin_port = htons(iPort);
	((SOCKADDR_IN *)siAddr)->sin_addr.s_addr = INADDR_ANY;

	//bind the socket to local address
	if(bind(*(sSocket), (SOCKADDR*)siAddr, sizeof(SOCKADDR_IN)) == SOCKET_ERROR)
	{
		return -1;
	}

	//wait for incoming connections on port
	if(listen(*(sSocket), MAX_CONNECTIONS) == SOCKET_ERROR)
	{
		return -1;
	}

	return 0;
}
